/*
 * Copyright 2014 Trend Micro Incorporated
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

#ifndef __TC_COMMON_H__
#define __TC_COMMON_H__
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <linux/version.h>

#include <net/pkt_sched.h>

enum link_layer
{
	LINKLAYER_UNSPEC,
	LINKLAYER_ETHERNET,
	LINKLAYER_ATM,
};

struct qdisc_util
{
	//struct  qdisc_util *next;
	const char *id;
	int	(*parse_qopt)(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n);
	//int	(*print_qopt)(struct qdisc_util *qu, FILE *f, struct rtattr *opt);
	//int 	(*print_xstats)(struct qdisc_util *qu, FILE *f, struct rtattr *xstats);

	int	(*parse_copt)(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n);
	//int	(*print_copt)(struct qdisc_util *qu, FILE *f, struct rtattr *opt);
};

struct filter_util
{
	//struct filter_util *next;
	char	id[16];
	int	(*parse_fopt)(struct filter_util *fu, char *fhandle, int argc,
			      char **argv, struct nlmsghdr *n);
	//int	(*print_fopt)(struct filter_util *qu, FILE *f, struct rtattr *opt, __u32 fhandle);
};

struct __tc_ratespec {
	unsigned char	cell_log;
	__u8		linklayer; /* lower 4 bits */
	unsigned short	overhead;
	short		cell_align;
	unsigned short	mpu;
	__u32		rate;
};

struct __tc_htb_opt {
	struct __tc_ratespec 	rate;
	struct __tc_ratespec 	ceil;
	__u32	buffer;
	__u32	cbuffer;
	__u32	quantum;
	__u32	level;		/* out only */
	__u32	prio;
};

struct nl_req
{
	struct nlmsghdr n;
	struct tcmsg t;
	char buf[0];
};

#define TCA_HTB_DIRECT_QLEN (TCA_HTB_RTAB + 1)

#define TC_LINKLAYER_MASK 0x0F

#define TCA_BUF_MAX	(64*1024)

#define MAX_MSG 16384

#define alloc_mem(_p, _size, _flags) \
	do \
	{ \
		_p = kmalloc((size_t)_size, _flags); \
		if (!_p) \
		{ \
			TC_ERR("Cannot allocate %zu bytes of memory from kernel.", (size_t)_size); \
			return -1; \
		} \
	} while (0)

#define free_mem(_p) \
	do \
	{ \
		if (_p) \
		{ \
			kfree(_p); \
			_p = NULL; \
		} \
	} while (0)

#define free_skb(_p) \
	do \
	{ \
		if (_p) \
		{ \
			kfree_skb(_p); \
			_p = NULL; \
		} \
	} while (0)

#define NLMSG_TAIL(nmsg) \
	((struct rtattr *) (((void *) (nmsg)) + NLMSG_ALIGN((nmsg)->nlmsg_len)))

#define NEXT_ARG() \
	do \
	{ \
		argc--; \
		if (argc <= 0) \
		{ \
			TC_ERR("Arguments required after \"%s\".", *argv); \
			return -1; \
		} \
		argv++; \
	} while (0)

#undef MEM_ERR
#define MEM_ERR(len) \
	do \
	{ \
		printk(KERN_ERR "[%s:%d] Failed to allocate %llu bytes of memory\n", __FUNCTION__, __LINE__, (uint64_t)len); \
	} while (0)

#undef TC_ERR
#define TC_ERR(fmt, args...) \
	do \
	{ \
		printk(KERN_ERR "[%s:%d] " fmt "\n", __FUNCTION__, __LINE__, ##args); \
	} while (0)

#undef TC_DBG
#define TC_DBG(fmt, args...) \
	do \
	{ \
		if (0) \
		{ \
			printk(KERN_INFO "[%s:%d] " fmt "\n", __FUNCTION__, __LINE__, ##args); \
		} \
	} while (0)


#define strtoul simple_strtoul
#define strtoull simple_strtoull
#define strtol simple_strtol

extern int matches(const char *cmd, const char *pattern);
extern int addattr_l(struct nlmsghdr *n, int maxlen, int type, const void *data, int alen);
extern struct rtattr *addattr_nest_compat(struct nlmsghdr *n, int maxlen, int type, const void *data, int len);
extern int addattr_nest_compat_end(struct nlmsghdr *n, struct rtattr *start);
extern int get_ifindex_by_name(const char *dev);
extern int get_hz(void);
extern int get_unsigned(unsigned *val, const char *arg, int base);
extern int get_u32(u32 *val, const char *arg, int base);
extern int get_u16(u16 *val, const char *arg, int base);
extern int get_integer(int *val, const char *arg, int base);
extern int get_size(unsigned *size, const char *str);
extern int get_size_and_cell(unsigned *size, int *cell_log, char *str);
extern int get_rate(unsigned *rate, const char *str);
extern int get_rate64(__u64 *rate, const char *str);
extern int tc_calc_rtable(struct __tc_ratespec *r, __u32 *rtab, int cell_log, unsigned mtu, enum link_layer linklayer);
extern unsigned tc_calc_xmittime(u32 rate, unsigned size);
//extern int tc_calc_xmittime(unsigned *value, u32 rate, unsigned size);
extern int get_handle(u32 *h, const char *str);
extern struct qdisc_util *get_qdisc_util(const char *str);
extern struct filter_util *get_filter_util(const char *str);
extern struct sk_buff *prepare_req_skb(int len, void *req);

// for HTB qdisc and class options
extern int htb_parse_opt(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n);
extern int htb_parse_class_opt(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n);

#ifdef PRIO_QDISC_SUPPORT
// for PRIO qdisc options
extern int prio_parse_opt(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n);
#endif

#ifdef FIFO_QDISC_SUPPORT
// for [p|b]fifo qdisc options
extern int fifo_parse_opt(struct qdisc_util *qu, int argc, char **argv, struct nlmsghdr *n);
#endif

// for u32 filter options
extern int u32_parse_filter_opt(struct filter_util *fu, char *handle, int argc, char **argv, struct nlmsghdr *n);

extern int do_qdisc(int argc, char **argv);
extern int do_class(int argc, char **argv);
extern int do_filter(int argc, char **argv);
extern int do_unicast(struct sk_buff *skb);
//extern int do_multicast(struct sk_buff *skb);

#endif	// __TC_COMMON_H__
